/*
Copyright 2009 Harteex Productions
License: Creative Commons Attribution 3.0 Unported
http://creativecommons.org/licenses/by/3.0/
*/

#include "mixer.h"

void mixerInit(unsigned int samplerate, unsigned short bits, unsigned char channels, unsigned char volume)
{
	sprintf(_lastError, "No Error");
	WaveoutOpen woOpen;
	woOpen.frequency = samplerate;
	woOpen.bits = bits;
	woOpen.channels = channels;
	woOpen.volume = volume;

	mixerVolume = volume;

	woHandle = waveout_open(&woOpen);

	mixerStopAllSound();
	
	curReadBuffer = 0;
	curWriteBuffer = 0;
	soundThreadExit = 0;
	soundThreadPause = 1;

	OSTaskCreate(_mixerPlay, (void *) 0, (void *)&TaskStartStk[TASK_STK_SIZE - 1], TASK_START_PRIO);
}

void mixerClose()
{
	soundThreadPause = 1;
	soundThreadExit = 1;
	mixerStopAllSound();

	while (soundThreadExit == 1);  // to avoid pb with waveout_close -> can hang ??

	mixerOggClear();
	mixerModClear();

	waveout_close(woHandle);
}

void mixerPause(bool b)
{
	soundThreadPause = b ? 1 : 0;
}

bool mixerIsPaused()
{
	return soundThreadPause;
}

void mixerVolumeIncrease()
{
	if (mixerVolume < VOLUME_MAX)
		mixerVolume += VOLUME_STEP;

	if (mixerVolume > VOLUME_MAX)
		mixerVolume = VOLUME_MAX;

	waveout_set_volume(mixerVolume);
}

void mixerVolumeDecrease()
{
	if (mixerVolume > 0)
		mixerVolume -= VOLUME_STEP;

	if (mixerVolume < 0)
		mixerVolume = 0;

	waveout_set_volume(mixerVolume);
}

int8_t mixerGetVolume()
{
	return mixerVolume;
}

void mixerSetVolume(int8_t v)
{
	if (v < 0)
		mixerVolume = 0;
	else if (v > VOLUME_MAX)
		mixerVolume = VOLUME_MAX;
	else
		mixerVolume = v;

	waveout_set_volume(mixerVolume);
}

void mixerSetSampleRate(uint16_t rate)
{
	pcm_ioctl(PCM_SET_SAMPLE_RATE, rate);
}

void mixerSetChannels(uint8_t channels)
{
	pcm_ioctl(PCM_SET_CHANNEL, channels);
}

void mixerSetBits(uint8_t bits)
{
	pcm_ioctl(PCM_SET_FORMAT, bits);
}

void mixerStopAllSound()
{
	memset(mixerSoundBufferTotal, 0, MAXFRAGSIZE * SOUNDBUFFERS);

	uint16_t i;
	for (i = 0; i < SOUNDBUFFERS; i++)
	{
		dataLength[i] = MAXFRAGSIZE;
	}

	mixerOggClear();
	mixerModClear();
}

void _mixerPlay(void *none)
{
	do
	{
		if (soundThreadPause == 0)
		{
			curReadBuffer++;
			if (curReadBuffer == SOUNDBUFFERS)
				curReadBuffer = 0;

			waveout_write(woHandle, mixerSoundBuffers[curReadBuffer], MAXFRAGSIZE);
		}
		else
		{
			OSTimeDly(1);
		}
	} while (soundThreadExit == 0); // Until the end of the sound thread

	soundThreadExit = 0;

	OSTaskDel(TASK_START_PRIO);
}

void mixerOggClear()
{
	oggEOF = true;
	mixerMusicEOF = true;
	oggBufferUnWritten = 0;

	if (oggFile != NULL)
	{
		ov_clear(oggFile);
		free(oggFile);
		oggFile = NULL;
	}

	if (oggFileHandle != NULL)
	{
		fsys_fclose(oggFileHandle);
		oggFileHandle = NULL;
	}

	songLength = 0;
	songCur = 0;
}

void mixerModClear()
{
	if (!modLoaded)
		return;

	mixerMusicEOF = true;

	MODFILE_Stop(modFile);
	MODFILE_Free(modFile);
	free(modFile);
	modLoaded = false;
}

bool mixerSetOgg(char* file)
{
	mixerTypePlaying = MIXER_TYPE_NONE;

	mixerOggClear();
	mixerModClear();

	oggFileHandle = fsys_fopen(file, "rb");
	if (oggFileHandle == NULL)
	{
		sprintf(_lastError, "Cannot open file. Bad file name?");
		return false;
	}

	oggFile = (OggVorbis_File*)malloc(sizeof(OggVorbis_File));
	if (oggFile == NULL)
	{
		fsys_fclose(oggFileHandle);
		oggFileHandle = NULL;
		sprintf(_lastError, "Out of memory!");
		return false;
	}

	if (ov_open(oggFileHandle, oggFile, NULL, 0) < 0)
	{
		// error
		free(oggFile);
		oggFile = NULL;
		fsys_fclose(oggFileHandle);
		oggFileHandle = NULL;
		sprintf(_lastError, "Decoder doesn't want to read the file. Invalid?");
		return false;
	}

	// Set correct rate and channels for this file
	oggInfo = ov_info(oggFile, -1);
	mixerSetChannels(oggInfo->channels);
	mixerSetSampleRate(oggInfo->rate);

	infoArtist[0] = 0;
	infoAlbum[0] = 0;
	infoTrack[0] = 0;
	vorbis_comment* tempComment = ov_comment(oggFile, -1);
	uint16_t i;
	for (i = 0; i < tempComment->comments; i++)
	{
		_parseOggComment(tempComment->user_comments[i], tempComment->comment_lengths[i]);
	}

	songLength = ov_time_total(oggFile, -1);

	oggEOF = false;
	mixerMusicEOF = false;
	mixerTypePlaying = MIXER_TYPE_OGG;
	return true;
}

void _parseOggComment(char* str, int length)
{
	void* p = memchr(str, '=', length);

	if (strncmpi("ARTIST=", str, 7) == 0)
	{
		strncpy(infoArtist, p + 1, min(length - 7, 255));
	}

	if (strncmpi("ALBUM=", str, 6) == 0)
	{
		strncpy(infoAlbum, p + 1, min(length - 6, 255));
	}

	if (strncmpi("TITLE=", str, 6) == 0)
	{
		strncpy(infoTrack, p + 1, min(length - 6, 255));
	}
}

bool mixerSetMod(char* file)
{
	mixerTypePlaying = MIXER_TYPE_NONE;

	mixerOggClear();
	mixerModClear();

	modFile = calloc(1, sizeof(MODFILE));
	if (modFile == NULL)
	{
		sprintf(_lastError, "Memory allocation failed");
		return false;
	}

	int ret = MODFILE_Load(file, modFile);
	if (ret < 0)
	{
		// error
		sprintf(_lastError, "Mod could not be loaded, error %i", ret);
		free(modFile);
		return false;
	}

	mixerSetChannels(2);
	mixerSetSampleRate(44100);

	modFile->musicvolume = 64;
	modFile->sfxvolume = 64;
	MODFILE_Start(modFile);
	MODFILE_SetFormat(modFile, 44100, 2, 16, 1);

	mixerTypePlaying = MIXER_TYPE_MOD;

	infoArtist[0] = 0;
	infoAlbum[0] = 0;
	infoTrack[0] = 0;
	strncpy(infoTrack, modFile->songname, 28);
	infoTrack[28] = 0;

	songLength = modFile->songlength;

	modLoaded = true;
	mixerMusicEOF = false;

	return true;
}

bool mixerUpdate()
{
	while (curWriteBuffer != curReadBuffer)
	{
		if (mixerTypePlaying == MIXER_TYPE_OGG)
		{
			_mixerUpdateOgg();
		}
		else if (mixerTypePlaying == MIXER_TYPE_MOD)
		{
			_mixerUpdateMod();
		}
		else
		{
			memset(mixerSoundBuffers[curWriteBuffer], 0, MAXFRAGSIZE);
		}

		curWriteBuffer++;
		curWriteBuffer = curWriteBuffer % SOUNDBUFFERS;
	}

	return !mixerMusicEOF;
}

void _mixerUpdateOgg()
{
	char* curBufferPtr = mixerSoundBuffers[curWriteBuffer];
	int curBufferLengthLeft = MAXFRAGSIZE;

	if (!oggEOF)
	{
		songCur = ov_time_tell(oggFile);

		bool curBufferDone = false;
		do {
			if (oggBufferUnWritten > 0)
			{
				// First fill with data from last read
				char * ptr = oggBuffer + oggBufferLastWriteLength - oggBufferUnWritten;
				memcpy(mixerSoundBuffers[curWriteBuffer], ptr, oggBufferUnWritten);

				curBufferPtr += oggBufferUnWritten;
				curBufferLengthLeft = MAXFRAGSIZE - oggBufferUnWritten;

				oggBufferUnWritten = 0;
				oggBufferLastWriteLength = 0;

				if (curBufferLengthLeft == 0)
					break;
			}

			long ret = ov_read(oggFile, oggBuffer, MAXFRAGSIZE, &oggCurrentSection);

			if (ret == 0)
			{
				// EOF
				oggEOF = true;
				mixerMusicEOF = true;
				curBufferDone = true;
			}
			else if (ret < 0)
			{
				// error in the stream.
				curBufferDone = true; // Just to not hang...
			}
			else
			{
				if (ret >= curBufferLengthLeft)
				{
					memcpy(curBufferPtr, oggBuffer, curBufferLengthLeft);
					oggBufferUnWritten = ret - curBufferLengthLeft;
					curBufferDone = true;
					oggBufferLastWriteLength = ret;
				}
				else
				{
					memcpy(curBufferPtr, oggBuffer, ret);
					curBufferPtr += ret;
					curBufferLengthLeft -= ret;
				}
			}
		} while (!curBufferDone);
	}
	
	if (oggEOF) // Not an else, since eof could be set above
	{
		if (curBufferLengthLeft > 0)
			memset(curBufferPtr, 0, curBufferLengthLeft);
	}
}

void _mixerUpdateMod()
{
	char* curBufferPtr = mixerSoundBuffers[curWriteBuffer];

	if (modLoaded && !mixerMusicEOF)
	{
		songCur = modFile->play_position;

		modFile->mixingbuf = (void*) curBufferPtr;
		modFile->mixingbuflen = MAXFRAGSIZE;
		MODFILE_Player(modFile);

		if (!modFile->playing)
			mixerMusicEOF = true;
	}
}

char* mixerLastError()
{
	return _lastError;
}

bool mixerSeek(int32_t ms)
{
	if (mixerTypePlaying == MIXER_TYPE_OGG)
	{
		if (oggEOF)
		{
			return false;
		}
		else
		{
			return ov_time_seek(oggFile, ms) == 0;
		}
	}

	return false;
}
